from __future__ import annotations

from copy import deepcopy
from typing import Any, cast

from django.core.exceptions import ImproperlyConfigured

from .base import Block


class DatagridItem(Block):
    """
    Implements a `Tabler datagrid-item
    <https://preview.tabler.io/docs/datagrid.html>`_ It should be used with
    :py:class:`Datagrid`.

    Note:
        Unlike :py:class:`wildewidgets.widgets.base.Block`,
        :py:class:`DatagridItem` requires either :py:attr:`contents` to be set,
        or the block contents to be provided as positional arguments.

    Examples:
        Create a :py:class:`DatagridItem` with a title and content:

        .. code-block:: python

            from wildewidgets import DatagridItem, Datagrid

            item = DatagridItem(
                "This is the content of the item.",
                title="My Item",
                url="https://example.com"
            )

            grid = Datagrid(item)

            # or

            grid = Datagrid()
            grid.add_item(item)


    Args:
        *blocks: strings or :py:class:`wildewidgets.widgets.base.Block` objects
            to add to our ``datagrid-item``.  If :py:attr:`contents` is not set,
            these blocks will be used as the contents of the ``datagrid-item``.

    Keyword Args:
        title: the ``datagrid-title`` of the ``datagrid-item``
        url: URL to use to turn content into a hyperlink

    Raises:
        ValueError: either the ``title`` was empty, or no contents were provided
        ImproperlyConfigured: :py:attr:`url` was set, and there is more than one block
            in :py:attr:`contents`

    """

    block: str = "datagrid-item"

    #: the ``datagrid-title`` of the ``datagrid-item``
    title: str | None = None
    #: a URL to use to turn our contents into a hyperlink
    url: str | None = None

    def __init__(
        self,
        *blocks: Any,
        title: str | None = None,
        url: str | None = None,
        **kwargs: Any,
    ) -> None:
        self.title = title or self.title
        assert self.title is not None, "title is required"  # noqa: S101
        if not self.title:
            msg = (
                '"title" is required as either a keyword argument or as a class '
                "attribute"
            )
            raise ValueError(msg)
        self.url = url or self.url
        super().__init__(*blocks, **kwargs)

    def add_blocks(self) -> None:
        """
        Add our content.

        If :py:attr:`url` is set, and there is only one block in
        :py:attr:`contents`, wrap that block in a :py:class:`wildewidgets.Link`.

        Raises:
            ImproperlyConfigured: :py:attr:`url` was set, and there is more than
                one block in :py:attr:`contents`

        """
        if self.url:
            if len(self.contents) == 1:
                wrapper: Block = Block(
                    self.contents[0],
                    tag="a",
                    attributes={"href": self.url},
                    name="datagrid-content",
                )
            else:
                msg = (
                    f"{self.__class__.__name__}: url should only be set if contents "
                    "are a single block"
                )
                raise ImproperlyConfigured(msg)
        elif len(self.contents) == 1:
            if not isinstance(self.contents[0], Block):
                # Wrap the text in a block to allow us to assign the datagrid-content
                # class to it
                wrapper = Block(self.contents[0], name="datagrid-content")
            else:
                wrapper = self.contents[0]
        else:
            wrapper = Block(name="datagrid-content")
            for block in self.contents:
                wrapper.add_block(block)
        self.add_block(Block(self.title, name="datagrid-title"))  # type: ignore[arg-type]
        self.add_block(wrapper)


DatagridItemDef = DatagridItem | tuple[str, str] | tuple[str, str, dict[str, Any]]


class Datagrid(Block):
    """
    Implements a `Tabler Data grid <https://preview.tabler.io/docs/datagrid.html>`_
    It contains :py:class:`DatagridItem` objects.

    Examples:
        Add :py:class:`DatagridItem` objects to this in one of these ways:

        As constructor arguments::

            >>> item1 = DatagridItem(title='foo', content='bar', url="https://example.com")
            >>> item2 = DatagridItem(title='baz', content='barney')
            >>> item3 = ['foo', 'bar']
            >>> grid = Datagrid(item1, item2, item3)

        By using :py:meth:`add_block` with a :py:class:`DatagridItem`:

            >>> grid = Datagrid(item1, item2, item3)
            >>> grid.add_block(DatagridItem(title='foo', content='bar'))

        By using :py:meth:`add_item`::

            >>> grid = Datagrid(item1, item2, item3)
            >>> grid.add_item('foo', 'bar')

    Args:
        *items: a list of ``datagrid-item`` definitions or
        *:py:class:`DatagridItem` objects.

    """

    block: str = "datagrid"
    #: a list of ``datagrid-items`` to add to our content
    items: list[DatagridItemDef] = []  # noqa: RUF012

    def __init__(self, *items: DatagridItemDef, **kwargs: Any) -> None:
        self.items = list(items) if items else deepcopy(self.items)
        super().__init__(**kwargs)
        for item in items:
            if isinstance(item, DatagridItem):
                self.add_block(item)
            elif isinstance(item, tuple):
                if len(item) == 2:  # noqa: PLR2004
                    item = cast("tuple[str, str]", item)
                    self.add_item(item[0], item[1])
                else:
                    item = cast("tuple[str, str, dict[str, Any]]", item)
                    self.add_item(item[0], item[1], **item[2])

    def add_item(
        self,
        title: str,
        content: str | Block | list[str | Block],
        url: str | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Add a :py:class:`DatagridItem` to our block contents, with
        ``datagrid-title`` of ``title`` and datagrid.

        Examples:
            Start with our grid::

                >>> dg = DataGrid()

            Add a simple key/value::

                >>> dg.add_item('Version', '1.2.3')

            Add a block as the value, and wrap it in a :py:class:`wildewdigets.Link`::

                >>> dg.add_item(
                    'Gravatar',
                    Image(src=static('myapp/images/gravatar.png'), alt='MyGravatar'),
                    url='https://www.google.com'
                )

            Add a list of blocks as the value::

                >>> dg.add_item(
                    'Contributors',
                    [
                        ImageLink(
                            src=static('myapp/images/fred-gravatar.png',
                            alt='Fred'
                            url='https://www.fred.com'
                        ),
                        ImageLink(
                            src=static('myapp/images/barney-gravatar.png',
                            alt='Barney'
                            url='https://www.barney.com'
                        )
                    ],
                    css_class='d-flex flex-row'
                )


        Note:
            To add a :py:class:`DatagridItem` directly, use :py:meth:`add_block`.

        Keyword Args:
            title: the ``datagrid-title`` of the ``datagrid-item``
            content: the ``datagrid-content`` of the ``datagrid-item``
            url: URL to use to turn content into a hyperlink
            **kwargs: additional keyword arguments passed to the
                :py:class:`DatagridItem` constructor

        """
        if isinstance(content, (str, Block)):
            content = [content]
        self.add_block(DatagridItem(*content, title=title, url=url, **kwargs))



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Any

from django.core.exceptions import ImproperlyConfigured
from django.urls import path, reverse_lazy
from django.utils.text import slugify

from wildewidgets.views.generic import (
    CreateView,
    DeleteView,
    IndexView,
    TableAJAXView,
    TableBulkActionView,
    UpdateView,
)
from wildewidgets.widgets import (
    BaseDataTable,
    BreadcrumbBlock,
    LookupModelTable,
    Navbar,
)

from .models import ViewSetMixin

if TYPE_CHECKING:
    from django.views import View


class ModelViewSet:
    """
    A viewset that provides CRUD operations for a Django model.

    This class simplifies the creation of a complete set of views for a model,
    automatically generating URLs and configuring views for listing, creating,
    updating, and deleting model instances. It follows the Django REST Framework
    ViewSet pattern but for traditional Django views.

    The viewset handles:

    - Index/list view with a data table
    - AJAX endpoint for table data loading
    - Bulk action processing (e.g., bulk delete)
    - Create view with a form
    - Update view with a form
    - Delete view

    To use this class, create a subclass and set at minimum the `model` attribute:

    Example:
        .. code-block:: python

            from wildewidgets import (
                BreadcrumbBlock,
                TablerVerticalNavbar,
                Menu,
                MenuItem,
            )
            from wildewidgets.viewsets import ModelViewSet
            from myapp.models import Book

            class MainMenu(Menu):
                items = [
                    MenuItem(
                        text="Home",
                        icon="house",
                        url=reverse_lazy("core:home"),
                    ),
                    MenuItem(
                        text="Books",
                        icon="mailbox",
                        url=reverse_lazy("core:books--index"),
                    ),
                ]

            class MyNavbar(TablerVerticalNavbar):
                hide_below_viewport: str = "xl"
                branding = Block(
                    LinkedImage(
                        image_src=static("images/logo.png"),
                        image_width="150px",
                        image_alt="MyOrg",
                        css_class="d-flex justify-content-center ms-3",
                        url="https://www.example.com/",
                    ),
                )
                contents = [
                    MainMenu(),
                ]


            class MyBreadcrumbs(BreadcrumbBlock):
                def __init__(self):
                    super().__init__()
                    self.add_breadcrumb(title="Home", url="/")

            class BookViewSet(ModelViewSet):
                model = Book
                url_prefix = "books"
                url_namespace = "myapp"
                breadcrumbs_class = MyBreadcrumbs

            # In urls.py
            urlpatterns += BookViewSet().get_urlpatterns()

    Keyword Args:
        model: The model class to use (must be a
            :py:class:`wildewidgets.ViewSetMixin` subclass)
        url_prefix: Optional prefix for all URLs (defaults to model name)
        url_namespace: Optional namespace for URL names
        **kwargs: Additional attributes to set on the viewset

    Raises:
        ImproperlyConfigured: If model is not provided or not a ViewSetMixin
            subclass, or if model is missing required attributes

    """

    #: The model for which we are building this viewset
    model: type[ViewSetMixin] | None = None

    #: The path prefix to use to organize our views.
    url_prefix: str | None = None
    #: The URL namespace to use for our view names.  This should be set to
    #: the ``app_name`` set in the ``urls.py`` where this viewset's patterns
    #: will be added to ``urlpatterns``
    url_namespace: str | None = None
    #: Set this to your :py:class:`wildewidgets.widgets.navigation.BreadcrumbBlock`
    #: class to manage breadcrumbs automatically.
    breadcrumbs_class: type[BreadcrumbBlock] | None = None
    #: The navbar to use.  We'll highlight our verbose name in any of the
    #: menus within the navbar
    navbar_class: type[Navbar] | None = None
    #: The template name to use for all views
    template_name: str | None = None

    #: The view class to use for the index view; must be a subclass of
    #: :py:class:`wildewidgets.views.generic.IndexView`.
    index_view_class: type[IndexView] = IndexView
    #: The :py:class:`wildewidgets.widgets.tables.base.BaseDataTable` subclass
    #: to use for the listing table
    index_table_class: type[BaseDataTable] = LookupModelTable
    #: A dictionary that we will use as the ``**kwargs`` for the constructor of
    #: :py:attr:`table_class`
    index_table_kwargs: dict[str, Any] = {  # noqa: RUF012
        "striped": True,
        "page_length": 50,
        "buttons": True,
    }

    #: The view class to use as the AJAX JSON endpoint for the dataTable; must
    #: be a subclass of :py:class:`wildewidgets.views.generic.TableAJAXView`.
    table_ajax_view_class: type[View] = TableAJAXView

    #: The view class to use for the index view; must be a subclass of
    #: :py:class:`wildewidgets.views.generic.CreateView`.
    table_bulk_action_view_class: type[View] = TableBulkActionView

    #: The view class to use for the index view; must be a subclass of
    #: :py:class:`wildewidgets.views.generic.CreateView`.
    create_view_class: type[View] = CreateView

    #: The view class to use for the index view; must be a subclass of
    #: :py:class:`wildewidgets.views.generic.UpdateView`.
    update_view_class: type[View] = UpdateView

    #: The view class to use for the index view; must be a subclass of
    #: :py:class:`wildewidgets.views.generic.DeleteView`.
    delete_view_class: type[View] = DeleteView

    def __init__(
        self,
        model: type[ViewSetMixin] | None = None,
        url_prefix: str | None = None,
        url_namespace: str | None = None,
        **kwargs,
    ):
        #: The model we're describing in this view. It must have
        #: :py:class:`wildewidgets.models.ViewSetMixin` in its class heirarchy
        self.model: type[ViewSetMixin] = model or self.model
        if not self.model or not issubclass(self.model, ViewSetMixin):
            msg = (
                "model must be either set as a class attribute or passed in as "
                "a kwarg, and it also must be a subclass of ViewSetMixin"
            )
            raise ImproperlyConfigured(msg)
        if not self.model._meta.verbose_name_plural:
            msg = "Model must have a verbose_name_plural set to use breadcrumbs"
            raise ImproperlyConfigured(msg)
        if not self.model._meta.verbose_name:
            msg = "model must have a verbose_name"
            raise ImproperlyConfigured(msg)
        # Tell the model we are its viewset
        self.model.viewset = self
        self.name = slugify(self.model._meta.verbose_name.lower())
        self.url_prefix = url_prefix or self.name
        self.url_namespace = url_namespace

        for key, value in kwargs.items():
            setattr(self, key, value)

    def get_url_name(self, view_name: str, namespace: bool = True) -> str:
        """
        Generate a URL name for a specific view.

        This method creates a URL name by combining the model name with the view name.
        If namespace is True and a URL namespace is set, the namespace is prepended.

        Args:
            view_name: The base name of the view (e.g., "index", "create")
            namespace: Whether to include the namespace prefix if available

        Returns:
            str: The complete URL name for the view

        Example:
            >>> viewset.get_url_name("create")
            'myapp:book--create'  # (if namespace is 'myapp' and model name is 'book')

        """
        url_name = f"{self.name}--{view_name}"
        if namespace and self.url_namespace:
            url_name = f"{self.url_namespace}:{url_name}"
        return url_name

    def get_breadcrumbs(self, view_name: str) -> BreadcrumbBlock | None:
        """
        Create breadcrumb navigation for the specified view.

        This method returns a configured breadcrumb block for the current view.
        For views other than "index", it adds a breadcrumb link back to the index view.

        Args:
            view_name: The name of the view ("index", "create", "update", etc.)

        Returns:
            Configured breadcrumb block, or None if :py:attr:`breadcrumbs_class`
            is not set

        """
        breadcrumbs: BreadcrumbBlock | None = None
        if self.breadcrumbs_class:
            breadcrumbs = self.breadcrumbs_class()  # pylint: disable=not-callable
            if view_name != "index":
                breadcrumbs.add_breadcrumb(
                    title=self.model.model_verbose_name_plural(),  # type: ignore[union-attr]
                    url=reverse_lazy(self.get_url_name("index")),
                )
        return breadcrumbs

    def get_index_table_kwargs(self) -> dict[str, Any]:
        """
        Get keyword arguments for configuring the index table.

        This method creates a dictionary of configuration options for the data table
        on the index page. It automatically sets:
        - Default form actions if none are specified
        - The AJAX URL for fetching data
        - The form URL for bulk actions

        Returns:
            dict[str, Any]: Configuration options for the data table

        """
        table_kwargs = deepcopy(self.index_table_kwargs)
        if "form_actions" not in table_kwargs:
            table_kwargs["form_actions"] = [("delete", "Delete selected")]
        table_kwargs["ajax_url_name"] = self.get_url_name("table-ajax")
        table_kwargs["form_url"] = reverse_lazy(self.get_url_name("table-bulk-actions"))
        return table_kwargs

    @property
    def index_view(self):
        """
        Get the configured index view.

        This property returns a Django view function for listing model instances
        in a data table, with all necessary configuration.

        Returns:
            View callable: Configured index view

        """
        return self.index_view_class.as_view(
            model=self.model,
            template_name=self.template_name,
            table_class=self.index_table_class,
            table_kwargs=self.get_index_table_kwargs(),
            breadcrumbs=self.get_breadcrumbs("index"),
            navbar_class=self.navbar_class,
        )

    @property
    def table_ajax_view(self):
        """
        Get the configured AJAX data endpoint view.

        This property returns a Django view function that serves as the AJAX endpoint
        for the data table, providing data in JSON format.

        Returns:
            View callable: Configured AJAX view

        """
        return self.table_ajax_view_class.as_view(
            model=self.model,
            table_class=self.index_table_class,
            table_kwargs=self.get_index_table_kwargs(),
        )

    @property
    def table_bulk_action_view(self):
        """
        Get the configured bulk action view.

        This property returns a Django view function for processing bulk actions
        on multiple model instances, such as bulk deletion.

        Returns:
            View callable: Configured bulk action view

        """
        return self.table_bulk_action_view_class.as_view(
            model=self.model,
            url=reverse_lazy(self.get_url_name("index")),
            table_class=self.index_table_class,
            table_kwargs=self.get_index_table_kwargs(),
        )

    @property
    def create_view(self):
        """
        Get the configured create view.

        This property returns a Django view function for creating new model instances,
        with form handling and success/error redirects.

        Returns:
            View callable: Configured create view

        """
        return self.create_view_class.as_view(
            model=self.model,
            template_name=self.template_name,
            breadcrumbs=self.get_breadcrumbs("create"),
            navbar_class=self.navbar_class,
            success_url=reverse_lazy(self.get_url_name("index")),
        )

    @property
    def update_view(self):
        """
        Get the configured update view.

        This property returns a Django view function for updating existing model
        instances, with form handling and success/error redirects.

        Returns:
            View callable: Configured update view

        """
        return self.update_view_class.as_view(
            model=self.model,
            template_name=self.template_name,
            breadcrumbs=self.get_breadcrumbs("update"),
            navbar_class=self.navbar_class,
            success_url=reverse_lazy(self.get_url_name("index")),
        )

    @property
    def delete_view(self):
        """
        Get the configured delete view.

        This property returns a Django view function for deleting model instances,
        with success redirect after deletion.

        Returns:
            View callable: Configured delete view

        """
        return self.delete_view_class.as_view(
            model=self.model, success_url=reverse_lazy(self.get_url_name("index"))
        )

    def get_urlpatterns(self):
        """
        Generate URL patterns for all views.

        This method creates a list of URL patterns for all CRUD operations:
        - Index/list view
        - AJAX endpoint for table data
        - Bulk action processing
        - Create view
        - Update view
        - Delete view

        The URLs follow a consistent pattern based on the model name and any prefix.

        Returns:
            list: Django URL patterns for all views

        Example:
            >>> urlpatterns = BookViewSet().get_urlpatterns()
            >>> # URLs will include patterns like 'books/', 'books/create/', etc.

        """
        root = f"{slugify(self.model.model_logger_name())}/"
        if self.url_prefix:
            root = f"{self.url_prefix}/{root}"
        return [
            path(
                root, self.index_view, name=self.get_url_name("index", namespace=False)
            ),
            path(
                f"{root}table/ajax/",
                self.table_ajax_view,
                name=self.get_url_name("table-ajax", namespace=False),
            ),
            path(
                f"{root}table/bulk/",
                self.table_bulk_action_view,
                name=self.get_url_name("table-bulk-actions", namespace=False),
            ),
            path(
                f"{root}create/",
                self.create_view,
                name=self.get_url_name("create", namespace=False),
            ),
            path(
                f"{root}<int:pk>/",
                self.update_view,
                name=self.get_url_name("update", namespace=False),
            ),
            path(
                f"{root}<int:pk>/delete/",
                self.delete_view,
                name=self.get_url_name("delete", namespace=False),
            ),
        ]



from django.contrib import admin
# from .models import SomeModel

# Register your models here to have them show up in the django admin.
# NOTE: This is not useful for access.caltech apps outside of dev, though, as django admin uses URLs that depend on
# the models' IDs, which doesn't fly with CIT_AUTH. So the django admin feature may just not be useful for an app.
# admin.site.register(SomeModel)



from __future__ import annotations

import json
import random
from typing import TYPE_CHECKING, Any, cast

from django import template
from django.http import JsonResponse

from wildewidgets.views import WidgetInitKwargsMixin

from ..base import Widget

if TYPE_CHECKING:
    from django.http.request import HttpRequest


class ApexDatasetBase(Widget):
    """
    Base class for Apex chart datasets.

    This class provides the foundation for defining datasets that can be added to
    ApexCharts. It handles the transformation of Python data into the format
    expected by the ApexCharts JavaScript library.

    Attributes:
        name: The name of the dataset, displayed in legends and tooltips
        chart_type: The specific chart type for this dataset (e.g., 'line', 'bar')
        data: The list of data points for this dataset

    Example:
        .. code-block:: python

            class MyDataset(ApexDatasetBase):
                name = "Monthly Sales"
                chart_type = "line"

                def __init__(self, **kwargs):
                    super().__init__(**kwargs)
                    self.data = [10, 41, 35, 51, 49, 62, 69, 91, 148]


        Or use subclassing to customize the dataset further:

        .. code-block:: python

            class CustomDataset(ApexDatasetBase):
                name = "Custom Data"
                chart_type = "bar"

                def __init__(self, **kwargs):
                    super().__init__(**kwargs)
                    self.data = self.load_custom_data()

                def load_custom_data(self) -> list[int]:
                    # Load or generate custom data here
                    return [5, 15, 25, 35, 45]

    """

    name: str | None = None
    chart_type: str | None = None

    def __init__(self, **kwargs: Any) -> None:  # noqa: ARG002
        """
        Initialize the dataset with empty data.

        Args:
            **kwargs: Additional keyword arguments (not used in base class)

        """
        self.data: list[Any] = []

    def get_options(self) -> dict[str, Any]:
        """
        Get the dataset options in the format expected by ApexCharts.

        This method transforms the dataset's properties into a dictionary format
        that can be serialized to JSON and used by the ApexCharts JavaScript library.

        Returns:
            dict: The dataset configuration options

        """
        options: dict[str, Any] = {"data": self.data}
        if self.name:
            options["name"] = self.name if self.name is not None else ""
        if self.chart_type:
            options["type"] = self.chart_type
        return options


class ApexJSONMixin:
    """
    A mixin class adding AJAX support to Apex Charts.

    This mixin enables Apex charts to load their data asynchronously via AJAX.
    It handles the HTTP request/response cycle and provides methods for
    preparing the chart data as JSON.

    Attributes:
        template_name: The Django template file for rendering the AJAX-enabled chart
        chart_options: Dictionary containing the chart configuration

    Note:
        Classes using this mixin should implement the `load` method to populate
        the chart's datasets.

    """

    template_name: str = "wildewidgets/apex_json.html"
    chart_options: dict[str, Any]

    def dispatch(self, request: HttpRequest, *args: Any, **kwargs: Any) -> JsonResponse:
        """
        Dispatch the request to the appropriate handler method.

        This method delegates to the appropriate HTTP method handler (e.g., get, post)
        based on the request method.

        Args:
            request: The HTTP request object
            *args: Additional positional arguments

        Keyword Arguments:
            **kwargs: Additional keyword arguments

        Returns:
            JsonResponse: The JSON response containing chart data

        """
        handler = getattr(self, cast("str", request.method).lower())
        return handler(request, *args, **kwargs)

    def get(self, _: HttpRequest, *args: Any, **kwargs: Any) -> JsonResponse:  # noqa: ARG002
        """
        Handle GET requests for chart data.

        This method is called when the client makes an AJAX GET request for
        chart data. It returns the chart's series data as JSON.

        Args:
            _: The HTTP request object (unused)
            *args: Additional positional arguments
            **kwargs: Additional keyword arguments

        Returns:
            JsonResponse: The JSON response containing chart data

        """
        data = self.get_series_data()
        return self.render_to_response(data)

    def render_to_response(
        self,
        context: dict[str, Any],
        **response_kwargs: Any,  # noqa: ARG002
    ) -> JsonResponse:
        """
        Render the chart data as a JSON response.

        Args:
            context: The chart data to be serialized as JSON

        Keyword Arguments:
            **response_kwargs: Additional keyword arguments (unused)

        Returns:
            JsonResponse: The JSON response containing chart data

        """
        return JsonResponse(context)

    def get_series_data(self, **kwargs: Any) -> dict[str, Any]:
        """
        Get the chart series data.

        This method loads the chart data and adds the series to the provided
        keyword arguments.

        Keyword Args:
            **kwargs: Keyword arguments to update with series data

        Returns:
            dict: The updated keyword arguments containing series data

        """
        self.load()
        kwargs["series"] = self.chart_options["series"]
        return kwargs

    def load(self) -> None:
        """
        Load datasets into the chart via AJAX.

        This method should be overridden by subclasses to populate the chart's
        datasets. It will be called when an AJAX request is made for chart data.

        Example:
            .. code-block:: python

                def load(self) -> None:
                    dataset = MyDataset()
                    dataset.data = [10, 41, 35, 51, 49, 62, 69, 91, 148]
                    self.add_dataset(dataset)

        """


class ApexChartBase(WidgetInitKwargsMixin):
    """
    Base class for Apex Charts in Django applications.

    This class provides the foundation for creating interactive Apex charts
    in Django applications. It handles chart configuration, dataset management,
    and rendering.

    Attributes:
        template_name: The Django template file for rendering the chart
        css_id: The HTML ID attribute for the chart container
        chart_type: The type of chart (e.g., 'line', 'bar', 'pie')
        chart_options: Dictionary containing all chart configuration options

    Example:
        .. code-block:: python

            class MyLineChart(ApexChartBase):
                chart_type = 'line'

                def __init__(self, **kwargs):
                    super().__init__(**kwargs)
                    dataset = MyDataset()
                    dataset.data = [10, 41, 35, 51, 49, 62, 69, 91, 148]
                    self.add_dataset(dataset)
                    self.add_categories(['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                                        'Jul', 'Aug', 'Sep'])

    """

    template_name: str = "wildewidgets/apex_chart.html"
    css_id: str | int | None = None
    chart_type: str | None = None

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the Apex chart.

        Args:
            *args: Positional arguments passed to parent class

        Keyword Args:
            **kwargs: Keyword arguments
                css_id: Optional HTML ID for the chart container

        """
        super().__init__(*args, **kwargs)
        self.chart_options: dict[str, Any] = {}
        self.chart_options["series"] = []
        self.css_id = kwargs.get("css_id", self.css_id)
        self.chart_options["chart"] = {}
        if self.chart_type:
            self.chart_options["chart"]["type"] = self.chart_type

    def add_dataset(self, dataset: ApexDatasetBase) -> None:
        """
        Add a dataset to the chart.

        This method adds a dataset to the chart's series. Each dataset represents
        a data series in the chart (e.g., a line in a line chart or a set of bars
        in a bar chart).

        Args:
            dataset: The dataset to add to the chart

        """
        self.chart_options["series"].append(dataset.get_options())

    def add_categories(self, categories: list[Any]) -> None:
        """
        Set the categories for the chart's x-axis.

        This method sets the labels for the x-axis categories. These are typically
        used for categorical data like months, product names, etc.

        Args:
            categories: List of category labels

        """
        if "xaxis" not in self.chart_options:
            self.chart_options["xaxis"] = {}
        self.chart_options["xaxis"]["categories"] = categories

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """
        Get the context data for rendering the chart template.

        This method prepares the data needed by the Django template to render
        the chart, including chart options, CSS ID, and extra data.

        Keyword Args:
            **kwargs: Additional context data

        Returns:
            dict: The context data for the template

        """
        kwargs["options"] = json.dumps(self.chart_options)
        if not self.css_id:
            self.css_id = str(random.randint(1, 100000))  # noqa: S311
        kwargs["css_id"] = self.css_id
        kwargs["wildewidgetclass"] = self.__class__.__name__
        kwargs["extra_data"] = self.get_encoded_extra_data()
        return kwargs

    def add_suboption(self, option: str, name: str, value: Any) -> None:
        """
        Add a nested configuration option to the chart.

        This method adds a configuration option to a nested section of the chart
        options. It's useful for setting options in sections like 'chart', 'xaxis',
        'yaxis', etc.

        Args:
            option: The parent option section (e.g., 'chart', 'xaxis')
            name: The name of the specific option
            value: The value to set for the option

        Example:
            .. code-block:: python

                # Set chart.height = 350
                chart.add_suboption('chart', 'height', 350)

                # Set tooltip.enabled = False
                chart.add_suboption('tooltip', 'enabled', False)

        """
        if option not in self.chart_options:
            self.chart_options[option] = {}
        self.chart_options[option][name] = value

    def __str__(self) -> str:
        """
        Return the HTML representation of the chart.

        This method allows the chart to be used directly in Django templates.

        Returns:
            str: The rendered HTML for the chart

        """
        return self.get_content()

    def get_content(self, **kwargs: Any) -> str:  # noqa: ARG002
        """
        Render the chart as HTML content.

        This method renders the chart using the specified template and context data.

        Args:
            **kwargs: Additional context data (unused)

        Returns:
            str: The rendered HTML content for the chart

        """
        context = self.get_context_data()
        html_template = template.loader.get_template(self.template_name)
        return html_template.render(context)


class ApexSparkline(ApexChartBase):
    """
    A specialized Apex chart for creating sparklines.

    Sparklines are small, simple, condensed charts typically shown inline with
    text or in a dashboard. This class configures an Apex chart with the
    appropriate settings for a sparkline visualization.

    Attributes:
        width: The width of the sparkline chart in pixels
        stroke_width: The width of the line/stroke in pixels

    Example:
        .. code-block:: python

            sparkline = ApexSparkline()
            dataset = MyDataset()
            dataset.data = [10, 41, 35, 51, 49, 62, 69, 91, 148]
            sparkline.add_dataset(dataset)

        Or use subclassing to customize the sparkline further:

        ... code-block:: python

            class CustomSparkline(ApexSparkline):
                width = 100
                stroke_width = 3

                def __init__(self, **kwargs):
                    super().__init__(**kwargs)
                    self.add_suboption("chart", "sparkline", {"enabled": True})
                    self.add_suboption("chart", "width", self.width)
                    self.add_suboption("stroke", "width", self.stroke_width)
                    self.add_dataset(self.get_dataset())

                def get_dataset(self) -> list[Any]:
                    # do your custom data loading here
                    return (
                        Model.objects
                        .order_by('datestamp')
                        .values_list('field_name', flat=True)
                    )

    """

    width: int = 65
    stroke_width: int = 2

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize the sparkline chart with appropriate options.

        Args:
            **kwargs: Keyword arguments passed to the parent class

        """
        super().__init__(**kwargs)
        self.add_suboption("chart", "sparkline", {"enabled": True})
        self.add_suboption("chart", "width", self.width)
        self.add_suboption("stroke", "width", self.stroke_width)
